"""uti_plot_com module containing plot utilities not specific to a particular backend

.. moduleauthor:: 
"""
from copy import *
from array import *
import traceback
#import sys
#import numpy as np

import uti_math
import uti_io

#****************************************************************************
def _multicolumn_file_load(fname): #MR31102017
    with open(fname, 'r') as f:
        header = f.readline().split(',')
        header = [x.replace('#', '').strip() for x in header]
    data = uti_io.read_ascii_data_cols(fname, '\t', _i_col_start=0, _i_col_end=-1, _n_line_skip=1)
    d = {}
    for i, k in enumerate(header):
        k_no_units = k.split('[')[0].strip()
        units = k.split('[')[1].split(']')[0].strip()
        d[k_no_units] = {
            'data': data[i],
            'label': k,
            'units': units,
        }
    # data, mode, allrange, arLabels, arUnits
    return d, None, [], [], []

#****************************************************************************
#OBSOLETE
def _traj_file_load(fname, traj_axis='x'):  #MR20160725
    nLinesHead = 1
    hlp = []
    with open(fname, 'r') as f:
        for i in range(nLinesHead):
            hlp.append(f.readline())

    data = uti_io.read_ascii_data_cols(fname, '\t', _i_col_start=0, _i_col_end=10, _n_line_skip=nLinesHead)

    mode = 3
    allrange = [
        data[5][0],  # z-coordinate begin
        data[5][-1],  # z-coordinate end
        len(data[0]),  # number of points
        min(data[1]),  # x-coordinate begin
        max(data[1]),  # x-coordinate end
        1,
        min(data[3]),  # y-coordinate begin
        max(data[3]),  # y-coordinate end
        1,
    ]
    arLabels = ['Longitudinal Position', 'Horizontal Position', 'Vertical Position']
    arUnits = ['m', 'm', 'm', 'm']
    if traj_axis == 'x':
        data = data[1]
        arLabels.append('Horizontal Coordinate')
    elif traj_axis == 'y':
        data = data[3]
        arLabels.append('Vertical Coordinate')
    else:
        raise ValueError('Parameter "axis" has wrong value: {}. Allowed values are "x" and "y"'.format(traj_axis))

    return data, mode, allrange, arLabels, arUnits

#****************************************************************************
#def srw_ascii_load(fname):
#def file_load(_fname, _read_labels=1):
def _file_load(_fname, _read_labels=1): #MR20160725
    nLinesHead = 11
    hlp = []
    #with open(_fname,'r') as f: hlp = f.readlines(nLinesHead)
    with open(_fname,'r') as f:
        for i in range(nLinesHead):
            hlp.append(f.readline())
  
    #ne, nx, ny, ns    = [   int(hlp[i].replace('#','').split()[0]) for i in [3,6,9,10] ]
    ne, nx, ny = [int(hlp[i].replace('#','').split()[0]) for i in [3,6,9]]
    ns = 1
    testStr = hlp[nLinesHead - 1]
    if testStr[0] == '#':
        ns = int(testStr.replace('#','').split()[0])
    else: nLinesHead -= 1

    e0,e1,x0,x1,y0,y1 = [float(hlp[i].replace('#','').split()[0]) for i in [1,2,4,5,7,8]]

    #data = np.squeeze(np.loadtxt(_fname, dtype=np.float64)) #get data from file (C-aligned flat)
    data = uti_io.read_ascii_data_cols(_fname, '\t', _i_col_start=0, _i_col_end=0, _n_line_skip=nLinesHead)[0]

    allrange = e0, e1, ne, x0, x1, nx, y0, y1, ny

    arLabels = ['Photon Energy', 'Horizontal Position', 'Vertical Position', 'Intensity']
    arUnits = ['eV', 'm', 'm', 'ph/s/.1%bw/mm^2']

    if _read_labels:

        arTokens = hlp[0].split(' [')
        arLabels[3] = arTokens[0].replace('#','')
        arUnits[3] = '';
        if len(arTokens) > 1:
            arUnits[3] = arTokens[1].split('] ')[0]
        #print(arLabels[3], arUnits[3])

        for i in range(3):
            arTokens = hlp[i*3 + 1].split()
            nTokens = len(arTokens)
            nTokensLabel = nTokens - 3
            nTokensLabel_mi_1 = nTokensLabel - 1
            strLabel = ''
            for j in range(nTokensLabel):
                strLabel += arTokens[j + 2]
                if j < nTokensLabel_mi_1: strLabel += ' '
            arLabels[i] = strLabel
            arUnits[i] = arTokens[nTokens - 1].replace('[','').replace(']','')

    m = _enum('T','V','H','E','HV','EV','EH','EHV')

    if ne==1 and nx==1 and ny==1 : mode = m.T
    if ne==1 and nx==1 and ny>1  : mode = m.V
    if ne==1 and nx>1  and ny==1 : mode = m.H
    if ne>1  and nx==1 and ny==1 : mode = m.E
    if ne==1 and nx>1  and ny>1  : mode = m.HV
    if ne>1  and nx==1 and ny>1  : mode = m.EV
    if ne>1  and nx>1  and ny==1 : mode = m.EH
    if ne>1  and nx>1  and ny>1  : mode = m.EHV

    #print(mode)
    return data, mode, allrange, arLabels, arUnits

#****************************************************************************
#def file_load(fname, read_labels=1, traj_report=False, traj_axis='x'): #MR20160729
#    if not traj_report:
#        return _file_load(fname, read_labels)
#    else:
#        return _traj_file_load(fname, traj_axis)

def file_load(fname, read_labels=1, multicolumn_data=False): #MR31102017
    if not multicolumn_data:
        return _file_load(fname, read_labels)
    else:
        return _multicolumn_file_load(fname)

#****************************************************************************
def rescale(maxabsval, strval):
    """Force labels to 1.0e-3 boundary which contains maxabsval
    :param double maxabsval: absolute value on axis
    :param str strval: units
    :return (multiplier, strval): axis multiplier, axis label
    """
    mult = 1
    if maxabsval>=1.0e2 and maxabsval<1.0e5:
        mult = 1.0e-3
        strval = 'k'+strval
    if maxabsval>=1.0e5 and maxabsval<1.0e8:
        mult = 1.0e-6
        strval = 'M'+strval
    if maxabsval>=1.0e8 and maxabsval<1.0e11:
        mult = 1.0e-6
        strval = 'G'+strval
    if maxabsval>=1.0e-4 and maxabsval<1.0e-1:
        mult = 1.0e3
        strval = 'm'+strval
    if maxabsval>=1.0e-7 and maxabsval<1.0e-4:
        mult = 1.0e6
        strval = 'u'+strval
    if maxabsval>=1.0e-10 and maxabsval<1.0e-7:
        mult = 1.0e9
        strval = 'n'+strval
    if maxabsval>1.0e-13 and maxabsval<1.0e-10:
        mult = 1.0e12
        strval = 'p'+strval
    return mult, strval

#****************************************************************************
def rescale_range(allrange, _ar_units, _ec=0, _xc=0, _yc=0):
    """Adjust graph axis ranges and labels to be 1.0e-3 boundary

    :param tuple allrange: Order of ranges: e, x, y
    :param tuple _ar_units: units for ranges [e, x, y]
    :param 
    """
    e0, e1, ne, x0, x1, nx, y0, y1, ny = allrange

    #em = abs(np.array([e0,e1])).max()
    #xm = abs(np.array([x0,x1])).max()
    #ym = abs(np.array([y0,y1])).max()

    abs_e0 = abs(e0); abs_e1 = abs(e1)
    em = abs_e0
    if(em < abs_e1): em = abs_e1

    abs_x0 = abs(x0); abs_x1 = abs(x1)
    xm = abs_x0
    if(xm < abs_x1): xm = abs_x1

    abs_y0 = abs(y0); abs_y1 = abs(y1)
    ym = abs_y0
    if(ym < abs_y1): ym = abs_y1
    
    #mult_e, str_e = _rescale(em,"eV")
    #mult_x, str_x = _rescale(xm,"m")
    #mult_y, str_y = _rescale(ym,"m")
    #mult_e, str_e = _rescale(em, _ar_units[0])
    #mult_x, str_x = _rescale(xm, _ar_units[1])
    #mult_y, str_y = _rescale(ym, _ar_units[2])
    mult_e, str_e = rescale(em, _ar_units[0])
    mult_x, str_x = rescale(xm, _ar_units[1])
    mult_y, str_y = rescale(ym, _ar_units[2])

    e0s = uti_math.num_round(e0*mult_e) #this is done to avoid in the labels numbers like "12.700000000001"
    e1s = uti_math.num_round(e1*mult_e)
    x0s = uti_math.num_round(x0*mult_x)
    x1s = uti_math.num_round(x1*mult_x)
    y0s = uti_math.num_round(y0*mult_y)
    y1s = uti_math.num_round(y1*mult_y)
    ecs = uti_math.num_round(_ec*mult_e)
    xcs = uti_math.num_round(_xc*mult_x)
    ycs = uti_math.num_round(_yc*mult_y)
    
    #allnewrange = e0*mult_e, e1*mult_e, ne, x0*mult_x, x1*mult_x, nx, y0*mult_y, y1*mult_y, ny, _ec*mult_e, _xc*mult_x, _yc*mult_y
    allnewrange = e0s, e1s, ne, x0s, x1s, nx, y0s, y1s, ny, ecs, xcs, ycs
    strval = str_e, str_x, str_y
    return allnewrange, strval
    
#****************************************************************************
def rescale_dim(_range, _base_unit=None):
    """Adjust range and units of a value ("dimension" of a plot) to be 1.0e-3 boundary

    :param list _range: min. and max. value of a range to be adjusted
    :param sting _base_unit: base unit (e.g. [m], [eV],...)
    :return: tuple containing new adjusted range and unit
    """

    #xm = abs(np.array([_range[0], _range[1]])).max()
    abs_x0 = abs(_range[0]); abs_x1 = abs(_range[1])
    xm = abs_x0
    if(xm < abs_x1): xm = abs_x1
    
    mult, unit = rescale(xm, _base_unit)
    newrange = deepcopy(_range)
    newrange[0] *= mult; newrange[1] *= mult
    return newrange, unit
 
#****************************************************************************
def _enum(*sequential, **named):
    enums = dict(zip(sequential, range(len(sequential))), **named)
    return type('Enum', (), enums)




# -*- coding: utf-8 -*-
#############################################################################
# SRW Brightness Calculation
#############################################################################

import math
from scipy import special
import numpy as np
from pykern import pkresource

#General constants
_ElMass_kg=9.10938e-31
_Elch=1.60217662e-19
_LightSp=299792458
_ElMass_MeV=0.5109989
_Planck_eVs=4.135667662e-15


#Load numerical arrays for universal functions
#fluxcorrectionarray = np.loadtxt("resource/gwSrwBrilUndHarmUnivFlux.txt")
#divcorrectionarray = np.loadtxt("resource/gwSrwBrilUndHarmUnivDiv.txt")
#sizecorrectionarray = np.loadtxt("resource/gwSrwBrilUndHarmUnivSize.txt")
fluxcorrectionarray = np.loadtxt(pkresource.filename("template/srw/brilliance/gwSrwBrilUndHarmUnivFlux.txt"))
divcorrectionarray = np.loadtxt(pkresource.filename("template/srw/brilliance/gwSrwBrilUndHarmUnivDiv.txt"))
sizecorrectionarray = np.loadtxt(pkresource.filename("template/srw/brilliance/gwSrwBrilUndHarmUnivSize.txt"))
#srwlib.srwl_uti_read_data_cols
#np.array(srwlib.srwl_uti_read_data_cols('gwSrwBrilUndHarmUnivFlux.txt', '\t'))
#srwl_uti_interp_2d(_x, _y, _x_min, _x_step, _nx, _y_min, _y_step, _ny, _ar_f, _ord=3, _ix_per=1, _ix_ofst=0)

#Undulator K and E functions
def getK(By,lam_u):
    """Return K value
    :param By: vertical magnetic field [T]
    :param lam_u: undulator period [m]
    """
    return (_Elch/2/math.pi/_ElMass_kg/_LightSp)*By*lam_u

def getE(nHarm, Ebeam, K, lam_u):
    """Return energy [eV] of nth harmonic
    :param nHarm: harmonic number
    :param Ebeam: electron beam energy [GeV]
    :param K: deflection parameter
    :param lam_u: undulator period [m]
    """
    gamma=1957*Ebeam
    return ((_Planck_eVs*_LightSp*nHarm*2*gamma*gamma)/(lam_u*(1+.5*K**2)))

def KtoE(K,E_elec,lam_u,n):
    #compute photon Energy (in KeV) from a given K value
    #E_elec: electron energy in GeV
    #lam_u: undulator wavelength in cm
    #n: harmonic number
    return 0.9496376*n*E_elec**2/lam_u/(1+K**2/2)


def CalcK(B,lam_u):
    #compute undulator K value from magnetic field B [T] and period lam_u [m]
    return 93.36*B*lam_u

def CalcFluxUnd(Ib,kx,kz,phix,phiz,n,nPer,enDetPar,relEnSpr):
    #flux function in SRW Igor Pro
    #returns flux [#photons/sec/0.1%bandwidth]
    #Ib: beam current in Amps
    #kx: horizontal undulator strength parameter
    #kz: vertical undulator strength parameter
    #phix: horizontal undulator phase parameter
    #phiz: vertical undulator phase parameter
    #n: harmonic number of undulator spectrum
    #nPer: number of periods in undulator magnetic array
    #enDetPar: relative difference from undulator energy from on resonance peak value dE/E
    #relEnSpr: energy spread of electron beam


    def JJbsfun(k12,k22,n):
        #same as srwBrilUndBessFactExt in SRW-Igor
        qq=(n/4.)*(k12-k22)/(1+0.5*(k12+k22))
        return (special.jv((n-1.)/2,qq)-special.jv((n+1.)/2,qq))**2+(k22/k12)*(special.jv((n-1.)/2,qq)+special.jv((n+1.)/2,qq))**2

    def srwBrilUndPhotEnDetunCor(dEperE, relEnSpr, K1e2, K2e2, n):
        #additional correction factor from detuning and energy spread... explanation for this?
        #refered to as G function in notes
        fit_width = 0.63276
        auxMult = n*n*(K1e2 + K2e2)/(1 + (K1e2 + K2e2)/2)/(fit_width*fit_width)
        a_sig = auxMult*2*relEnSpr
        a_sigE2d2 = a_sig*a_sig/2
        v = 1 - special.erf(math.sqrt(a_sigE2d2))
        genFact = 0.0 if v == 0.0 else 0.5 + 0.5*v*math.exp(a_sigE2d2)
        if dEperE >= 0:
            res = genFact
        else:
            relarg = auxMult*dEperE
            res = math.exp(relArg)*genFact
        return res

    C0=4.5546497e13 #convConstFlux = alpha dw/w /e (dw/w = 0.001)
    #n=harmNum
    N=nPer
    normDetun = n*N*enDetPar
    normEnSpr = n*N*relEnSpr
    ke2=kx**2+kz**2
    phix=0 #what is phix?
    phiz=0 #what is phiz?
    phi0=0.5*math.atan((kz**2)*math.sin(2*phix)+(kx**2)*math.sin(2*phiz)/((kz**2)*math.cos(2*phix)+(kx**2)*math.cos(2*phiz)))
    k12=(kz**2)*(math.cos(phix-phi0))**2+(kx**2)*(math.cos(phiz-phi0))**2
    k22=(kz**2)*(math.sin(phix-phi0))**2+(kx**2)*(math.sin(phiz-phi0))**2
    JJbs=JJbsfun(k12,k22,n)
    #now get additional factors from energy spread and detuning
    #factDetunAndEnSpr = math.pi/2 #assumes zero detuning and energy spread, needs to be replaced with interpolation of external correction array
    factDetunAndEnSpr = interpBright(normDetun,normEnSpr,fluxcorrectionarray,-10,0,0.033389,0.02512565,600,200)
    GG=srwBrilUndPhotEnDetunCor(relEnSpr, enDetPar, k12, k22, n)

    return C0*N*Ib*(n*k12/(1+ke2/2))*JJbs*factDetunAndEnSpr*GG

def srwl_und_flux_en(Ib,kxmax,kzmax,kmin,numkpts,E_elec,lam_u,phix,phiz,n,nPer,enDetPar,relEnSpr):
    #compute kvals and Evals
    #lam_u: undulator wavelength in cm
    kmax = math.sqrt(kxmax**2+kzmax**2)
    dk = (kmax - kmin)/numkpts
    kvals=np.arange(kmin, kmax,dk)
    #compute Evals
    Evals = KtoE(kvals,E_elec,lam_u,n)
    #compute kxvals and kzvals
    if kxmax > kmin:
        dkx = (kxmax-kmin)/numkpts
        kxvals = np.arange(kmin,kxmax,dkx)
    else:
        kxvals = np.zeros(numkpts)

    if kzmax > kmin:
        dkz = (kzmax-kmin)/numkpts
        kzvals = np.arange(kmin,kzmax,dkz)
    else:
        kzvals = np.zeros(numkpts)

    #compute flux for each k value
    fluxvals = []
    for j in range(len(kvals)):
        #print j
        fluxvals.append(CalcFluxUnd(Ib,kxvals[j],kzvals[j],0,0,n,nPer,enDetPar,relEnSpr))
    return (Evals,fluxvals)


def CalcSizeUnd(sigsq,L,K,E_elec,lam_u,n,nPer,enDetPar,relEnSpr):
    #sigsq: hor. or vert. RMS electron beamsize squared [m^2]
    #L: length of undulator [m]
    #K: K value of undulator
    #E_elec: Energy of electron beam [GeV]
    #n: harmonic number
    #lam_u: undulator wavelength in cm
    normDetun = n*nPer*enDetPar
    normEnSpr = n*nPer*relEnSpr
    convConstSize = 0.5*1.239842e-06*L
    energy = 1000*KtoE(K,E_elec,lam_u,n)
    invSqrt2=1/math.sqrt(2)
    factAngDivDetunAndEnSpr = interpBright(normDetun,normEnSpr,sizecorrectionarray,-10,0,0.033389,0.02512565,600,200)*invSqrt2
    return math.sqrt(sigsq + (convConstSize/energy)*factAngDivDetunAndEnSpr**2)

def srwl_und_size_en(kxmax,kzmax,kmin,numkpts,E_elec,lam_u,phix,phiz,n,nPer,enDetPar,relEnSpr,sigsq):
    #compute kvals and Evals
    #lam_u: undulator wavelength in cm
    kmax = math.sqrt(kxmax**2+kzmax**2)
    dk = (kmax - kmin)/numkpts
    kvals=np.arange(kmin, kmax,dk)
    #compute Evals
    Evals = KtoE(kvals,E_elec,lam_u,n)
    #compute kxvals and kzvals
    if kxmax > kmin:
        dkx = (kxmax-kmin)/numkpts
        kxvals = np.arange(kmin,kxmax,dkx)
    else:
        kxvals = np.zeros(numkpts)

    if kzmax > kmin:
        dkz = (kzmax-kmin)/numkpts
        kzvals = np.arange(kmin,kzmax,dkz)
    else:
        kzvals = np.zeros(numkpts)

    L=(lam_u/100.)*nPer
    #compute size for each k value
    sizevals = []
    for j in range(len(kvals)):
        sizevals.append(CalcSizeUnd(sigsq,L,kvals[j],E_elec,lam_u,n,nPer,enDetPar,relEnSpr))
    return (Evals,sizevals)

def CalcDivergenceUnd(sigpsq,L,K,E_elec,lam_u,n,nPer,enDetPar,relEnSpr):
    #sigpsq: hor. or vert. RMS electron divergence squared
    #L: length of undulator [m]
    #K: K value of undulator
    #E_elec: Energy of electron beam [GeV]
    #n: harmonic number
    #lam_u: undulator wavelength in cm
    normDetun = n*nPer*enDetPar
    normEnSpr = n*nPer*relEnSpr
    convConstDiv = 2*1.239842e-06/L
    energy = 1000*KtoE(K,E_elec,lam_u,n)
    invSqrt2=1/math.sqrt(2)
    factAngDivDetunAndEnSpr = interpBright(normDetun,normEnSpr,divcorrectionarray,-10,0,0.033389,0.02512565,600,200)*invSqrt2
    return math.sqrt(sigpsq + (convConstDiv/energy)*factAngDivDetunAndEnSpr**2)

def srwl_und_div_en(kxmax,kzmax,kmin,numkpts,E_elec,lam_u,phix,phiz,n,nPer,enDetPar,relEnSpr,sigpsq):
    #compute kvals and Evals
    #lam_u: undulator wavelength in cm
    kmax = math.sqrt(kxmax**2+kzmax**2)
    dk = (kmax - kmin)/numkpts
    kvals=np.arange(kmin, kmax,dk)
    #compute Evals
    Evals = KtoE(kvals,E_elec,lam_u,n)
    #compute kxvals and kzvals
    if kxmax > kmin:
        dkx = (kxmax-kmin)/numkpts
        kxvals = np.arange(kmin,kxmax,dkx)
    else:
        kxvals = np.zeros(numkpts)

    if kzmax > kmin:
        dkz = (kzmax-kmin)/numkpts
        kzvals = np.arange(kmin,kzmax,dkz)
    else:
        kzvals = np.zeros(numkpts)

    L=(lam_u/100.)*nPer
    #compute divergence for each k value
    divergevals = []
    for j in range(len(kvals)):
        divergevals.append(CalcDivergenceUnd(sigpsq,L,kvals[j],E_elec,lam_u,n,nPer,enDetPar,relEnSpr))
    return (Evals,divergevals)


def CalcAngularfluxUnd(Ib,kx,kz,phix,phiz,n,nPer,E_elec,lam_u,enDetPar,relEnSpr,sigpxsq,sigpzsq):
    #Ib: beam current in Amps
    #kx: horizontal undulator strength parameter
    #kz: vertical undulator strength parameter
    #phix: horizontal undulator phase parameter
    #phiz: vertical undulator phase parameter
    #n: harmonic number of undulator spectrum
    #nPer: number of periods in undulator magnetic array
    #enDetPar: relative difference from undulator energy from on resonance peak value dE/E
    #relEnSpr: energy spread of electron beam
    #lam_u: undulator wavelength in cm
    L=(lam_u/100.)*nPer
    convConstDiv = 2*1.239842e-06/L
    K=math.sqrt(kx**2+kz**2)
    flux = CalcFluxUnd(Ib,kx,kz,phix,phiz,n,nPer,enDetPar,relEnSpr)
    divx = CalcDivergenceUnd(sigpxsq,L,K,E_elec,lam_u,n,nPer,enDetPar,relEnSpr)
    divz = CalcDivergenceUnd(sigpzsq,L,K,E_elec,lam_u,n,nPer,enDetPar,relEnSpr)
    fluxdivide = (2e+06*math.pi)*divx*divz
    return flux/fluxdivide

def srwl_und_ang_flux_en(Ib,kxmax,kzmax,kmin,numkpts,E_elec,lam_u,phix,phiz,n,nPer,enDetPar,relEnSpr,sigpxsq,sigpzsq):
     #compute kvals and Evals
     #lam_u: undulator wavelength in cm
    kmax = math.sqrt(kxmax**2+kzmax**2)
    dk = (kmax - kmin)/numkpts
    kvals=np.arange(kmin, kmax,dk)
    #compute Evals
    Evals = KtoE(kvals,E_elec,lam_u,n)
    #compute kxvals and kzvals
    if kxmax > kmin:
        dkx = (kxmax-kmin)/numkpts
        kxvals = np.arange(kmin,kxmax,dkx)
    else:
        kxvals = np.zeros(numkpts)

    if kzmax > kmin:
        dkz = (kzmax-kmin)/numkpts
        kzvals = np.arange(kmin,kzmax,dkz)
    else:
        kzvals = np.zeros(numkpts)

    L=(lam_u/100.)*nPer
    #compute flux for each k value
    angularflux = []
    for j in range(len(kvals)):
        angularflux.append(CalcAngularfluxUnd(Ib,kxvals[j],kzvals[j],phix,phiz,n,nPer,E_elec,lam_u,enDetPar,relEnSpr,sigpxsq,sigpzsq))
    return (Evals,angularflux)


def CalcBrightnessUnd(Ib,kx,kz,phix,phiz,n,E_elec,lam_u,nPer,enDetPar,relEnSpr,L,sigxsq,sigzsq,sigxpsq,sigzpsq):
    #compute Brightness from undulator by dividing flux by Sigx*Sigx'*Sigz*Sigz', with Sigx,z and Sigz,z'
    #photon beamsize at center of undulator
    #uses same parameters as srwl_compute_flux, plus
    #
    #L: undulator length
    #sigx: RMS horizontal electron beam size
    #sigz: RMS vertical electron beam size
    #sigxp: RMS horizontal electron beam divergence
    #sigzp: RMS vertical electron beam divergence
    cst=(math.pi*2)**2*1e12

    flux = CalcFluxUnd(Ib,kx,kz,phix,phiz,n,nPer,enDetPar,relEnSpr)

    K = math.sqrt(kx**2+kz**2)

    Sigmax = CalcSizeUnd(sigxsq,L,K,E_elec,lam_u,n,nPer,enDetPar,relEnSpr)
    Sigmaz = CalcSizeUnd(sigzsq,L,K,E_elec,lam_u,n,nPer,enDetPar,relEnSpr)
    Sigmaxp = CalcDivergenceUnd(sigxpsq,L,K,E_elec,lam_u,n,nPer,enDetPar,relEnSpr)
    Sigmazp = CalcDivergenceUnd(sigzpsq,L,K,E_elec,lam_u,n,nPer,enDetPar,relEnSpr)

    return flux/(cst*Sigmax*Sigmaxp*Sigmaz*Sigmazp)

def srwl_und_bright_en(Ib,kx,kz,phix,phiz,n,E_elec,lam_u,nPer,epeak,enDetPar,relEnSpr,L,sigxsq,sigzsq,sigxpsq,sigzpsq,kxmax,kzmax,kmin,numkpts):
    #compute kvals and Evals
    #lam_u: undulator wavelength in cm
    kmax = math.sqrt(kxmax**2+kzmax**2)
    dk = (kmax - kmin)/numkpts
    kvals=np.arange(kmin, kmax,dk)
    #compute Evals
    Evals = KtoE(kvals,E_elec,lam_u,n)
    #compute kxvals and kzvals
    if kxmax > kmin:
        dkx = (kxmax-kmin)/numkpts
        kxvals = np.arange(kmin,kxmax,dkx)
    else:
        kxvals = np.zeros(numkpts)

    if kzmax > kmin:
        dkz = (kzmax-kmin)/numkpts
        kzvals = np.arange(kmin,kzmax,dkz)
    else:
        kzvals = np.zeros(numkpts)

    brightnessvals = []
    for j in range(len(kvals)):
           brightnessvals.append(CalcBrightnessUnd(Ib,kxvals[j],kzvals[j],phix,phiz,n,E_elec,lam_u,nPer,enDetPar,relEnSpr,L,sigxsq,sigzsq,sigxpsq,sigzpsq))
    return (Evals,brightnessvals)

#*********************************************************************


#Analytic formula for calculation of beam size with detuning
def srwl_und_size_en_fixedK(sigsq,L,K,E_elec,lam_u,n,nPer,epeak,emin,emax,numepts, relEnSpr):
    #compute evals and detuned size vals

    evals = np.arange(emin,emax,(emax-emin)/numepts)
    enDetPars = (evals-epeak)/epeak

    #print(enDetPars)
     #compute size for each E value
    sizedet = []
    for j in range(len(enDetPars)):
        sizedet.append(CalcSizeUnd(sigsq,L,K,E_elec,lam_u,n,nPer,enDetPars[j],relEnSpr))

    return (evals,sizedet)


#Analytic formula for calculation of beam divergence with detuning
def srwl_und_div_en_fixedK(sigsq,L,K,E_elec,lam_u,n,nPer,epeak,emin,emax,numepts, relEnSpr):
    #compute evals and detuned div vals

    evals = np.arange(emin,emax,(emax-emin)/numepts)
    enDetPars = (evals-epeak)/epeak

    #print(enDetPars)
     #compute div for each E value
    divdet = []
    for j in range(len(enDetPars)):
        divdet.append(CalcDivergenceUnd(sigsq,L,K,E_elec,lam_u,n,nPer,enDetPars[j],relEnSpr))

    return (evals,divdet)


#Analytic formula for calculation of flux with detuning
def srwl_und_flux_en_fixedK(Ib,kx,kz,E_elec,lam_u,phix,phiz,n,nPer,epeak,emin,emax,numepts, relEnSpr):
    #compute evals and flux vals

    evals = np.arange(emin,emax,(emax-emin)/numepts)
    enDetPars = (evals-epeak)/epeak

     #compute flux for each E value
    fluxvals = []
    for j in range(len(enDetPars)):
        fluxvals.append(CalcFluxUnd(Ib,kx,kz,0,0,n,nPer,enDetPars[j],relEnSpr))

    return (evals,fluxvals)

#Analytic formula for calculation of brightness with detuning
def srwl_und_bright_en_fixedK(Ib,kx,kz,phix,phiz,n,E_elec,lam_u,nPer,epeak,emin,emax,numepts,relEnSpr,L,sigxsq,sigysq,sigxpsq,sigypsq):
    #compute evals and bright vals

    brightevals = np.arange(emin,emax,(emax-emin)/numepts)
    enDetPars = (brightevals-epeak)/epeak

    #print(enDetPars)
    #compute brightness for each E value
    brightdet = []

    for j in range(len(enDetPars)):
        brightdet.append(CalcBrightnessUnd(Ib,kx,kz,phix,phiz,n,E_elec,lam_u,nPer,enDetPars[j],relEnSpr,L,sigxsq,sigysq,sigxpsq,sigypsq))

    return (brightevals,brightdet)



#Interpolation function for universal functions
def interpBright(x,y,W,xmin,ymin,xstep,ystep,nx,ny):
    #interpolate the function in the array W.
    #x and y are the coordinates to evaluate it.
    #W is a 2-D array.  We need the minimum x and y, xmin and ymin,
    #and also the spacing xstep and ystep to find the correct values in W.
    #With nx and ny, we can quickly check if the requested value is out
    #of bounds, in which case we find the closest boundary value.

    xmax = xmin + (nx - 1)*xstep
    ymax = ymin + (ny - 1)*ystep

    #if target point is outside of range put it on boundary
    if(x<xmin):
        x=xmin
    if(y<ymin):
        y=ymin
    if(x>=xmax):
        x = xmax - xstep
    if(y>=ymax):
        y = ymax - ystep

    #now find surrounding integers for (x,y)

    [djx,jx0]=np.modf((x-xmin)/xstep)
    [djy,jy0]=np.modf((y-ymin)/ystep)
    jx0=int(jx0)
    jy0=int(jy0)

    #now get values

    W00 = W[jx0,jy0]
    W01 = W[jx0,jy0+1]
    W10 = W[jx0+1,jy0]
    #W11 = W[jx0+1,jy0+1]

    return W00 +  djx*(W01-W00) + djy*(W10-W00) #+ djx*djy*W11



"""Simple 1D & 2D plotting utilities package for "Synchrotron Radiation Workshop" (SRW).

``uti_plot`` currently wraps ``matplotlib``, but other backends are
planned.  If no suitable backend is available, ``uti_plot_init`` sets
the backend to ``uti_plot_none`` so that the calling program is still
functional.  This is useful for systems where there is no graphing
library available, but you still want to see the results of the
SRW program.

Usage:

    import uti_plot as up

    up.uti_plot_init()
    uti_plot1d(...)
    uti_plot_show()

Modules:

    uti_plot
        This module, which loads all other modules dynamically

    uti_plot_matplotlib
        Does the actually plotting using matplotlib.pyplot.  Currently, loaded in all cases except
        when ``backend`` is ``None``

    test_uti_plot
        Simple tests for uti_plot

.. moduleauthor:: Rob Nagler <nagler@radiasoft.net>
"""
import uti_plot_com
import sys
import traceback

_backend = None

DEFAULT_BACKEND = '<default>'

def uti_plot_init(backend=DEFAULT_BACKEND, fname_format=None):
    """Initializes plotting engine with backend and, optionally, save plots to fname_format

    Tries to initialize `backend` as the plotting engine.  If not found, an
    error will be printed, and this module's functions will be no-ops.  If
    DEFAULT_BACKEND provided, an appropriate backend will be chosen and printed.
    Plots may also be saved if fname_format is supplied.

    You may call ``uti_plot_init(None)`` explicitly so that no plotting occurs.

    :param str backend: a matplot backend (TkAgg, etc.) or ``inline`` in IPython
    :param str fname_format: where to save plots. format field is a sequential plot number, starting at 0.
    """
    global _backend
    if backend is not None:
        try:
            import uti_plot_matplotlib
            _backend = uti_plot_matplotlib.Backend(backend, fname_format)
            return
        except:
            traceback.print_exc()
            print(backend + ': unable to import specified backend (or its dependency); no plots')
    elif fname_format is not None:
        #raise Value(fname_format + ': fname_format must be null if backend is None')
        raise ValueError(fname_format + ': fname_format must be null if backend is None')
    _backend = _BackendNone()

def uti_plot_show():
    """Display the plots"""
    #if '_backend' not in locals(): uti_plot_init() #?
    _backend.uti_plot_show()

def uti_plot1d(ar1d, x_range, labels=('Photon Energy [eV]', 'ph/s/0.1%bw'), units=None):
    """Generate one-dimensional line plot from given array

    :param array ar1d: data points
    :param list x_range: Passed to numpy.linspace(start sequence, stop sequnce, num samples)
    :param tuple labels: [x-axis, y-axis]
    """
    #if '_backend' not in locals(): uti_plot_init() #?
    
    if(units is not None):
        x_range, x_unit = uti_plot_com.rescale_dim(x_range, units[0])
        units = [x_unit, units[1]]
        strTitle = '' if(len(labels) < 3) else labels[2]
        labels = (labels[0] + ' [' + units[0] + ']', labels[1] + ' [' + units[1] + ']', strTitle)

    #print('uti_plot1d:', x_range)
    _backend.uti_plot1d(ar1d, x_range, labels)

def uti_plot1d_ir(ary, arx=None, labels=('Longitudinal Position [m]', 'Horizontal Position [m]'), units=None): #OC25102019
#def uti_plot1d_ir(ary, arx, labels=('Longitudinal Position [m]', 'Horizontal Position [m]'), units=None): #OC15112017
    """Generate one-dimensional line plot from given array

    :param array arx: abscissa array
    :param array ary: ordinate array
    :param tuple labels: [x-axis, y-axis]
    """
    #if '_backend' not in locals(): uti_plot_init() #?
    
    if(units is not None):
        #x_range = [min(arx), max(arx), len(arx)]
        #x_range, x_unit = uti_plot_com.rescale_dim(x_range, units[0])
        #units = [x_unit, units[1]]
        strTitle = '' if(len(labels) < 3) else labels[2]
        labels = (labels[0] + ' [' + units[0] + ']', labels[1] + ' [' + units[1] + ']', strTitle)

    #OC25102019 (added stuff below)
    if(arx is None):
        ary0 = ary
        if(not(isinstance(ary0, list) or isinstance(ary0, array) or isinstance(ary0, tuple))):
            raise Exception("Incorrect definition of data arrays / lists to be plotted")

        lenData = len(ary)
        #arx = array.array('d', [0]*lenData)
        #aryNew = array.array('d', [0]*lenData)
        arx = [0]*lenData
        aryNew = [0]*lenData
        for i in range(lenData):
            ay_i = ary[i]
            arx[i] = ay_i[0]
            aryNew[i] = ay_i[1]
        ary = aryNew
    
    _backend.uti_plot1d_ir(ary, arx, labels)

def uti_plot1d_m(ars, labels=('X', 'Y'), units=None, styles=None, legend=None): #OC25102019
    """Plot multiple one-dimensional curves in one graph

    :param array ars: multiple data arrays, assuming: ars = [curve_1,curve_2,...], where curve_i can be:
        curve_i = [[y1_i,y2_i,...],[x_min_i,x_max_i,nx_i]] or
        curve_i = [[y1_i,y2_i,...],[x1_i,x2_i,...]] or
        curve_i = [[x1_i,y1_i],[x2_i,y2_i],...] 
    :param tuple labels: (x-axis, y-axis)
    :param tuple units: (x-units, y-units)
    """
    #if '_backend' not in locals(): uti_plot_init() #?
    if(units is not None):
        strTitle = '' if(len(labels) < 3) else labels[2]
        labels = (labels[0] + ' [' + units[0] + ']', labels[1] + ' [' + units[1] + ']', strTitle)

    _backend.uti_plot1d_m(ars, labels, styles, legend)

def uti_plot2d(ar2d, x_range=None, y_range=None, labels=('Horizontal Position [m]','Vertical Position [m]'), units=None): #OC30052020
#def uti_plot2d(ar2d, x_range, y_range, labels=('Horizontal Position [m]','Vertical Position [m]'), units=None):
    """Generate quad mesh plot from given "flattened" array

    :param array ar2d: flat 2d array or list of data points, non-flat 2d array or list, or PIL image
    :param list x_range: Passed to numpy.linspace(start sequence, stop sequnce, num samples)
    :param list y_range: y axis (same structure as x_range)
    :param tuple labels: [x-axis, y-axis]
    """
    #if '_backend' not in locals(): uti_plot_init() #?

    if((units is not None) and (x_range is not None) and (y_range is not None)): #OC30052020
    #if(units is not None):
        x_range, x_unit = uti_plot_com.rescale_dim(x_range, units[0])
        y_range, y_unit = uti_plot_com.rescale_dim(y_range, units[1])
        units = [x_unit, y_unit,  units[2]]
        strTitle = '' if(len(labels) < 3) else labels[2]
        labels = (labels[0] + ' [' + units[0]+ ']', labels[1] + ' [' + units[1] + ']', strTitle)

    _backend.uti_plot2d(ar2d, x_range, y_range, labels)

def uti_plot2d1d(ar2d, x_range, y_range, x=0, y=0, labels=('Horizontal Position', 'Vertical Position', 'Intensity'), units=None, graphs_joined=True):
    """Generate 2d quad mesh plot from given "flattened" array, and 1d cuts passing through (x, y)

    :param array ar2d: flat 2d array of data points
    :param list x_range: Passed to numpy.linspace(start sequence, stop sequnce, num samples)
    :param list y_range: y axis (same structure as x_range)
    :param x: x value for 1d cut
    :param y: y value for 1d cut
    :param tuple labels: [x-axis, y-axis, z-axis]
    :param tuple units: [x-axis, y-axis, z-axis]
    :param graphs_joined: switch specifying whether the 2d plot and 1d cuts have to be displayed in one panel or separately
    """
    
    #if '_backend' not in locals(): uti_plot_init() #?
    if(units is not None): #checking / re-scaling x, y

        #OC17032019
        xRangeOrig = x_range[1] - x_range[0]
        yRangeOrig = y_range[1] - y_range[0] #AH30082019
        #yStartOrig = y_range[1] - y_range[0]
        
        x_range, x_unit = uti_plot_com.rescale_dim(x_range, units[0])
        y_range, y_unit = uti_plot_com.rescale_dim(y_range, units[1])

        #OC17032019
        if(x != 0): x *= (x_range[1] - x_range[0])/xRangeOrig
        if(y != 0): y *= (y_range[1] - y_range[0])/yRangeOrig
        
        units = [x_unit, y_unit,  units[2]]

        strTitle = labels[2]
        label2D = (labels[0] + ' [' + units[0]+ ']', labels[1] + ' [' + units[1] + ']', strTitle)

        #strTitle = 'At ' + labels[1] + ': ' + str(y)
        strTitle = 'At ' + labels[1] + ': ' + str(round(y, 6)) #OC17032019
        
        if y != 0: strTitle += ' ' + units[1]
        label1X = (labels[0] + ' [' + units[0] + ']', labels[2] + ' [' + units[2] + ']', strTitle)

        #strTitle = 'At ' + labels[0] + ': ' + str(x)
        strTitle = 'At ' + labels[0] + ': ' + str(round(x, 6)) #OC17032019
        
        if x != 0: strTitle += ' ' + units[0]
        label1Y = (labels[1] + ' [' + units[1] + ']', labels[2] + ' [' + units[2] + ']', strTitle)
        
    else: #OC081115
        strTitle = labels[2]
        label2D = (labels[0], labels[1], strTitle)

        strTitle = 'At ' + labels[1] + ': ' + str(y)
        label1X = (labels[0], labels[2], strTitle)

        strTitle = 'At ' + labels[0] + ': ' + str(x)
        label1Y = (labels[1], labels[2], strTitle)

    labels = [label2D, label1X, label1Y]

    _backend.uti_plot2d1d(ar2d, x_range, y_range, x, y, labels, graphs_joined)

#def uti_plot_img(_img, _x_range=None, _y_range=None, ): #OC29052020
#    _backend.uti_plot_img(_img)

def uti_plot_data_file(_fname, _read_labels=1, _e=0, _x=0, _y=0, _graphs_joined=True, #Same as uti_data_file_plot, but better fits function name decoration rules in this module (uti_plot*)
                       _multicolumn_data=False, _column_x=None, _column_y=None, #MR31102017
                       _scale='linear', _width_pixels=None):
    """Generate plot from configuration in _fname

    :param str _fname: config loaded from here
    :param bool _read_labels: whether to read labels from _fname
    :param float _e: photon energy adjustment
    :param float _x: horizonal position adjustment
    :param float _y: vertical position adjustment
    :param bool _graphs_joined: if true, all plots in a single figure
    :param bool _multicolumn_data: if true, visualize multicolumn data data
    :param str _column_x: column for horizontal axis
    :param str _column_x: column for vertical axis
    :param str _scale: the scale to use for plotting data (linear by default, but could use log, log2, log10)  
    :param int _width_pixels: the width of the final plot in pixels  
    """
    #if '_backend' not in locals(): uti_plot_init() #?
    _backend.uti_plot_data_file(_fname, _read_labels, _e, _x, _y, _graphs_joined,
                                _multicolumn_data, _column_x, _column_y, #MR31102017
                                _scale, _width_pixels)

#def uti_data_file_plot(_fname, _read_labels=1, _e=0, _x=0, _y=0, _graphs_joined=True):
#def uti_data_file_plot(_fname, _read_labels=1, _e=0, _x=0, _y=0, _graphs_joined=True, _traj_report=False, _traj_axis='x'): #MR29072016
#def uti_data_file_plot(_fname, _read_labels=1, _e=0, _x=0, _y=0, _graphs_joined=True, _traj_report=False, _traj_axis='x', _scale='linear', _width_pixels=None): #MR20012017  
def uti_data_file_plot(_fname, _read_labels=1, _e=0, _x=0, _y=0, _graphs_joined=True,
                       _multicolumn_data=False, _column_x=None, _column_y=None, #MR31102017
                       _scale='linear', _width_pixels=None):
    """Generate plot from configuration in _fname

    :param str _fname: config loaded from here
    :param bool _read_labels: whether to read labels from _fname
    :param float _e: photon energy adjustment
    :param float _x: horizonal position adjustment
    :param float _y: vertical position adjustment
    :param bool _graphs_joined: if true, all plots in a single figure
    :param bool _multicolumn_data: if true, visualize multicolumn data data
    :param str _column_x: column for horizontal axis
    :param str _column_x: column for vertical axis
    :param str _scale: the scale to use for plotting data (linear by default, but could use log, log2, log10)  
    :param int _width_pixels: the width of the final plot in pixels  
    """
    #if '_backend' not in locals(): uti_plot_init() #?
    #_backend.uti_data_file_plot(_fname, _read_labels, _e, _x, _y, _graphs_joined)
    #_backend.uti_data_file_plot(_fname, _read_labels, _e, _x, _y, _graphs_joined, _traj_report, _traj_axis) #MR29072016
    #_backend.uti_data_file_plot(_fname, _read_labels, _e, _x, _y, _graphs_joined, _traj_report, _traj_axis, _scale, _width_pixels) #MR20012017  
    #_backend.uti_data_file_plot(_fname, _read_labels, _e, _x, _y, _graphs_joined,
    #                            _multicolumn_data, _column_x, _column_y, #MR31102017
    #                            _scale, _width_pixels)
    uti_plot_data_file(_fname, _read_labels, _e, _x, _y, _graphs_joined, _multicolumn_data, _column_x, _column_y, _scale, _width_pixels) #OC16112017

class _BackendBase(object):
    def __getattr__(self, attr):
        return self._backend_call

class _BackendMissing(_BackendBase):
    def _backend_call(self, *args, **kwargs):
        uti_plot_init()
        method_name = sys._getframe(1).f_code.co_name
        func = getattr(_backend, method_name)
        return func(*args)

class _BackendNone(_BackendBase):
    def _backend_call(*args, **kwargs):
        pass

_backend = _BackendMissing()
